"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = __importDefault(require("lodash"));
const fp_1 = __importDefault(require("lodash/fp"));
const geo_1 = require("./geo");
const dates_1 = require("./dates");
const string_1 = require("./string");
const ip_1 = require("./ip");
const p = __importStar(require("../../parser"));
const interfaces_1 = require("../../interfaces");
function buildLogicFn(parser, typeConfig = {}) {
    return walkAst(parser.ast, typeConfig);
}
exports.default = buildLogicFn;
function makeGetFn(field) {
    if (!field)
        return fp_1.default.values;
    if (field.includes('.'))
        return fp_1.default.get(field);
    return function getProp(obj) {
        return obj[field];
    };
}
function logicNode(field, cb) {
    if (field && string_1.isWildCardString(field)) {
        return string_1.findWildcardField(field, cb);
    }
    const get = makeGetFn(field);
    const getAnyData = makeSomeFn(cb);
    return function _logicNode(obj) {
        const data = get(obj);
        if (Array.isArray(data)) {
            return getAnyData(data);
        }
        return cb(data);
    };
}
function makeSomeFn(cb) {
    return function someFn(data) {
        return data.some(cb);
    };
}
function makeNegate(fn) {
    return function negate(data) {
        return !fn(data);
    };
}
const rangeMapping = {
    gte: lodash_1.default.gte,
    gt: lodash_1.default.gt,
    lte: lodash_1.default.lte,
    lt: lodash_1.default.lt
};
function rangeFn(node) {
    const { left, right } = node;
    if (!right) {
        return function singleRangeTerm(data) {
            return rangeMapping[left.operator](data, left.value);
        };
    }
    return function doubleRangeTerm(data) {
        return rangeMapping[left.operator](data, left.value)
            && rangeMapping[right.operator](data, right.value);
    };
}
function valueExists(value) {
    return value != null;
}
function isFalse() {
    return false;
}
function walkAst(node, typeConfig) {
    if (p.isEmptyAST(node)) {
        return isFalse;
    }
    if (p.isNegation(node)) {
        const childLogic = walkAst(node.node, typeConfig);
        return makeNegate(childLogic);
    }
    if (p.isGroupLike(node)) {
        return makeGroupFn(node, typeConfig);
    }
    const value = p.getAnyValue(node);
    const field = p.getField(node);
    if (p.isTerm(node)) {
        return logicNode(field, typeFunctions(node, typeConfig, makeIsValue(value)));
    }
    if (p.isRange(node)) {
        return logicNode(field, typeFunctions(node, typeConfig, rangeFn(node)));
    }
    if (p.isFunctionExpression(node)) {
        return logicNode(field, node.instance.match);
    }
    if (p.isGeoDistance(node)) {
        return logicNode(field, geo_1.geoDistance(node));
    }
    if (p.isGeoBoundingBox(node)) {
        return logicNode(field, geo_1.geoBoundingBox(node));
    }
    if (p.isExists(node)) {
        return logicNode(field, valueExists);
    }
    if (p.isRegexp(node)) {
        return logicNode(field, string_1.regexp(node.value));
    }
    if (p.isWildcard(node)) {
        return logicNode(field, string_1.wildcard(node.value));
    }
    // nothing was found so return a fn that returns false
    return isFalse;
}
function typeFunctions(node, typeConfig, defaultCb) {
    if (node.field == null)
        return defaultCb;
    const type = typeConfig[node.field];
    if (type === interfaces_1.FieldType.Date) {
        if (p.isRange(node)) {
            return dates_1.dateRange(node);
        }
        return dates_1.compareTermDates(node);
    }
    if (type === interfaces_1.FieldType.IP) {
        if (p.isRange(node)) {
            return ip_1.ipRange(node);
        }
        return ip_1.ipTerm(node);
    }
    return defaultCb;
}
function makeIsValue(value) {
    return function isValue(data) {
        return data === value;
    };
}
function makeConjunctionFn(conjunction, typeConfig) {
    const fns = conjunction.nodes.map((node) => walkAst(node, typeConfig));
    return makeAllPassFn(fns);
}
function makeGroupFn(node, typeConfig) {
    const fns = node.flow.map((conjunction) => makeConjunctionFn(conjunction, typeConfig));
    return makeAnyPassFn(fns);
}
function makeAnyPassFn(fns) {
    const len = fns.length;
    return function anyPassFn(data) {
        for (let i = 0; i < len; i++) {
            if (fns[i](data))
                return true;
        }
        return false;
    };
}
function makeAllPassFn(fns) {
    const len = fns.length;
    return function allPassFn(data) {
        for (let i = 0; i < len; i++) {
            if (!fns[i](data))
                return false;
        }
        return true;
    };
}
//# sourceMappingURL=index.js.map