"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@terascope/utils");
const util_1 = require("util");
const utils_2 = require("../utils");
const i = __importStar(require("./interfaces"));
const utils = __importStar(require("./utils"));
const functions_1 = __importDefault(require("./functions"));
const interfaces_1 = require("../interfaces");
function makeContext(args) {
    let typeConfig;
    let variables;
    let logger;
    // eslint-disable-next-line
    ({ typeConfig = {}, variables = {}, logger } = args);
    if (!typeConfig || !logger) {
        throw new Error('Peg Engine given invalid context');
    }
    /**
    * Propagate the default field on a field group expression
    */
    function propagateDefaultField(node, field) {
        if (!node)
            return;
        if (utils.isTermType(node) && !node.field) {
            node.field = field;
            coerceTermType(node);
            return;
        }
        if (utils.isNegation(node)) {
            propagateDefaultField(node.node, field);
            return;
        }
        if (utils.isGroupLike(node)) {
            for (const conj of node.flow) {
                propagateDefaultField(conj, field);
            }
            return;
        }
        if (utils.isConjunction(node)) {
            for (const conj of node.nodes) {
                propagateDefaultField(conj, field);
            }
        }
    }
    // The following functions are javascript functions and not
    // peg grammar because this is based off configuration passed
    // in and cannot be inferred by the syntax
    function getFieldType(field) {
        if (!field)
            return;
        return typeConfig[field];
    }
    function getVariable(value) {
        const variable = variables[value];
        if (variable === undefined)
            throw new utils_1.TSError(`Could not find a variable set with key "${value}"`);
        if (Array.isArray(variable))
            return variable.slice();
        return variable;
    }
    const inferredFieldTypes = [interfaces_1.FieldType.String];
    function isInferredTermType(field) {
        const fieldType = getFieldType(field);
        if (!fieldType)
            return false;
        return inferredFieldTypes.includes(fieldType);
    }
    // parse an inferred field type
    function parseInferredTermType(field, value) {
        const fieldType = getFieldType(field);
        const term = {
            type: i.ASTType.Term,
            field_type: fieldType,
        };
        if (fieldType === interfaces_1.FieldType.String) {
            term.quoted = false;
            term.value = `${value}`;
            return term;
        }
        logger.warn(`Unsupported field inferred field type ${fieldType} for field ${field}`);
        term.value = value;
        return term;
    }
    function parseFunction(field, name, params) {
        const fnType = functions_1.default[name];
        if (fnType == null)
            throw new Error(`Could not find an xlucene function with name "${name}"`);
        // we are delaying instantiation until after parser since this can be called multiple times
        return () => fnType.create(field, params, { logger, typeConfig });
    }
    function makeFlow(field, values, varName) {
        return values.map((value) => {
            validateRestrictedVariable(value, varName);
            const node = { field, type: i.ASTType.Term, value };
            coerceTermType(node);
            // this creates an OR statement
            return {
                type: i.ASTType.Conjunction,
                nodes: [node]
            };
        });
    }
    // cannot allow variables that are objects, errors, maps, sets, buffers
    function validateRestrictedVariable(variable, varName) {
        const type = utils_1.getTypeOf(variable);
        if (!['String', 'Number', 'Boolean', 'RegExp'].includes(type)) {
            throw new Error(`Unsupported type of ${type} received for variable $${varName}`);
        }
    }
    function coerceTermType(node, _field) {
        if (!node)
            return;
        const field = node.field || _field;
        if (!field)
            return;
        const fieldType = getFieldType(field);
        if (fieldType === node.field_type)
            return;
        logger.trace(`coercing field "${field}":${node.value} type of ${node.field_type} to ${fieldType}`);
        // in the case of tokenized fields we should update the
        // node to indicate so non-term level queries can be performed
        if (utils.isTermType(node) && field.includes('.') && !typeConfig[field]) {
            const parentField = field.split('.').slice(0, -1);
            if (typeConfig[parentField] && typeConfig[parentField] !== interfaces_1.FieldType.Object) {
                node.tokenizer = parentField;
            }
        }
        if (fieldType === interfaces_1.FieldType.Boolean) {
            node.field_type = fieldType;
            delete node.quoted;
            delete node.restricted;
            if (node.value === 'true') {
                node.value = true;
            }
            if (node.value === 'false') {
                node.value = false;
            }
            return;
        }
        if (fieldType === interfaces_1.FieldType.Integer) {
            if (utils.isRange(node)) {
                node.left.field_type = fieldType;
                if (node.right)
                    node.right.field_type = fieldType;
            }
            else {
                delete node.quoted;
                delete node.restricted;
                node.field_type = fieldType;
                node.value = parseInt(node.value, 10);
            }
            return;
        }
        if (fieldType === interfaces_1.FieldType.Float) {
            node.field_type = fieldType;
            delete node.quoted;
            delete node.restricted;
            node.value = parseFloat(node.value);
            return;
        }
        if (fieldType === interfaces_1.FieldType.String) {
            node.field_type = fieldType;
            if (util_1.isRegExp(node.value)) {
                node.type = i.ASTType.Regexp;
                node.value = node.value.source;
            }
            else {
                node.quoted = false;
                node.value = `${node.value}`;
            }
        }
    }
    return {
        logger,
        parseGeoPoint: utils_2.parseGeoPoint,
        parseGeoDistance: utils_2.parseGeoDistance,
        coerceTermType,
        parseInferredTermType,
        isInferredTermType,
        propagateDefaultField,
        parseFunction,
        getVariable,
        makeFlow,
        validateRestrictedVariable
    };
}
exports.default = makeContext;
//# sourceMappingURL=context.js.map