"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@terascope/utils");
const helpers_1 = require("./helpers");
const utils_2 = require("../../../utils");
const interfaces_1 = require("../../../translator/interfaces");
const interfaces_2 = require("../../../interfaces");
const string_1 = require("../../../document-matcher/logic-builder/string");
const compatMapping = {
    [interfaces_2.GeoShapeType.Polygon]: interfaces_1.ESGeoShapeType.Polygon,
    [interfaces_2.GeoShapeType.MultiPolygon]: interfaces_1.ESGeoShapeType.MultiPolygon,
};
function validate(params) {
    const geoPointsParam = params.find((node) => node.field === 'points');
    const geoRelationParam = params.find((node) => node.field === 'relation');
    let relation;
    if (geoRelationParam) {
        const relationKeys = Object.values(interfaces_2.GeoShapeRelation);
        if (!relationKeys.includes(geoRelationParam.value)) {
            throw new utils_1.TSError(`Invalid relation value "${geoRelationParam.value}"`);
        }
        relation = geoRelationParam.value;
    }
    else {
        relation = interfaces_2.GeoShapeRelation.Within;
    }
    if (geoPointsParam == null)
        throw new utils_1.TSError('Invalid geoPolygon query, need to specify a "points" parameter');
    let polygonShape = {
        type: interfaces_2.GeoShapeType.Polygon,
        coordinates: []
    };
    if (helpers_1.isGeoShapePolygon(geoPointsParam.value) || helpers_1.isGeoShapeMultiPolygon(geoPointsParam.value)) {
        polygonShape = geoPointsParam.value;
    }
    else {
        if (!Array.isArray(geoPointsParam.value))
            throw new utils_1.TSError('Invalid points parameter, it must either be a geoshape or be an array of geo-points');
        const points = geoPointsParam.value.map((node) => {
            const value = node.value || node;
            const { lat, lon } = utils_2.parseGeoPoint(value);
            return [lon, lat];
        });
        const coords = helpers_1.validateListCoords(points);
        polygonShape.coordinates = coords;
    }
    return { polygonShape, relation };
}
const geoPolygon = {
    name: 'geoPolygon',
    version: '1',
    // @ts-ignore type issues with esPolyToPointQuery AnyQuery results
    create(_field, params, { logger, typeConfig }) {
        if (!_field || _field === '*')
            throw new Error('Field for geoPolygon cannot be empty or "*"');
        const { polygonShape, relation } = validate(params);
        let type;
        if (string_1.isWildCardString(_field)) {
            const regex = string_1.parseWildCard(_field);
            const results = [];
            // collect all pertinent typeConfig fields to wildcard
            for (const [key] of Object.entries(typeConfig)) {
                if (string_1.matchString(key, regex))
                    results.push(typeConfig[key]);
            }
            const types = utils_1.uniq(results);
            if (types.length > 1)
                throw new utils_1.TSError(`Cannot query geoPolygon against different field types ${JSON.stringify(types)}`);
            [type] = types;
        }
        else {
            type = typeConfig[_field];
            // can remove the second check when "geo" if fully deprecated
        }
        const targetIsGeoPoint = type === interfaces_2.FieldType.GeoPoint
            || type === interfaces_2.FieldType.Geo
            || type === undefined;
        function ESPolyQuery(field, points) {
            return {
                geo_polygon: {
                    [field]: {
                        points
                    }
                }
            };
        }
        function makePolygonQuery(field, coordinates) {
            // it has no holes
            if (coordinates.length === 1) {
                return ESPolyQuery(field, coordinates[0]);
            }
            const query = {
                bool: {
                    should: []
                }
            };
            const filter = {
                bool: {
                    filter: []
                }
            };
            const holes = {
                bool: {
                    must_not: []
                }
            };
            coordinates.forEach((coords, index) => {
                if (index === 0) {
                    filter.bool.filter.push(ESPolyQuery(field, coords));
                }
                else {
                    holes.bool.must_not.push(ESPolyQuery(field, coords));
                }
            });
            filter.bool.filter.push(holes);
            query.bool.should.push(filter);
            return query;
        }
        function esPolyToPointQuery(field) {
            // TODO: chech if points is a polygon with holes
            if (helpers_1.isGeoShapePolygon(polygonShape)) {
                const query = makePolygonQuery(field, polygonShape.coordinates);
                if (logger.level() === 10)
                    logger.trace('built geo polygon to point query', { query });
                return { query };
            }
            if (helpers_1.isGeoShapeMultiPolygon(polygonShape)) {
                const query = {
                    bool: {
                        should: polygonShape.coordinates.map((polyCoords) => makePolygonQuery(field, polyCoords))
                    }
                };
                if (logger.level() === 10)
                    logger.trace('built geo polygon to point query', { query });
                return { query };
            }
            return { query: {} };
        }
        function esPolyToPolyQuery(field) {
            const esType = compatMapping[polygonShape.type] || interfaces_1.ESGeoShapeType.Polygon;
            // @ts-ignore
            const query = {
                geo_shape: {
                    [field]: {
                        shape: {
                            type: esType,
                            coordinates: polygonShape.coordinates
                        },
                        relation
                    }
                }
            };
            if (logger.level() === 10)
                logger.trace('built geo polygon to polygon query', { query });
            return { query };
        }
        function polyToGeoPointMatcher() {
            const polygon = helpers_1.makeShape(polygonShape);
            // Nothing matches so return false
            if (polygon == null)
                return () => false;
            return helpers_1.polyHasPoint(polygon);
        }
        function polyToGeoShapeMatcher() {
            const polygon = helpers_1.makeShape(polygonShape);
            if (polygon == null)
                return () => false;
            const fn = helpers_1.polyHasShape(polygon, relation);
            return fn;
        }
        return {
            match: targetIsGeoPoint ? polyToGeoPointMatcher() : polyToGeoShapeMatcher(),
            toElasticsearchQuery: targetIsGeoPoint ? esPolyToPointQuery : esPolyToPolyQuery
        };
    }
};
exports.default = geoPolygon;
//# sourceMappingURL=polygon.js.map