"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const ts = __importStar(require("@terascope/utils"));
const p = __importStar(require("../parser"));
const translator_1 = require("../translator");
const interfaces_1 = require("../interfaces");
const string_1 = require("../document-matcher/logic-builder/string");
const _logger = ts.debugLogger('xlucene-query-access');
class QueryAccess {
    constructor(config = {}, options = {}) {
        var _a, _b;
        this._parser = new p.CachedParser();
        this._translator = new translator_1.CachedTranslator();
        const { excludes = [], includes = [], constraint, allow_empty_queries: allowEmpty = true, } = config;
        const typeConfig = config.type_config || options.type_config || {};
        const variables = options.variables || {};
        if (ts.isEmpty(typeConfig))
            throw new Error('Configuration for type_config must be provided');
        this.typeConfig = { ...typeConfig };
        this.logger = options.logger != null
            ? options.logger.child({ module: 'xlucene-query-access' })
            : _logger;
        this.excludes = (_a = excludes) === null || _a === void 0 ? void 0 : _a.slice();
        this.includes = (_b = includes) === null || _b === void 0 ? void 0 : _b.slice();
        this.constraints = ts.castArray(constraint).filter(Boolean);
        this.allowEmpty = Boolean(allowEmpty);
        this.preventPrefixWildcard = Boolean(config.prevent_prefix_wildcard);
        this.allowImplicitQueries = Boolean(config.allow_implicit_queries);
        this.defaultGeoField = config.default_geo_field;
        this.defaultGeoSortOrder = config.default_geo_sort_order;
        this.defaultGeoSortUnit = config.default_geo_sort_unit;
        this.parsedTypeConfig = this._restrictTypeConfig();
        this.variables = variables;
    }
    clearCache() {
        this._parser.reset();
        this._translator.reset();
    }
    /**
     * Validate and restrict a xlucene query
     *
     * @returns a restricted xlucene query
     */
    restrict(q, options = {}) {
        return this._restrict(q, options).query;
    }
    /**
     * Validate and restrict a xlucene query
     *
     * @returns a restricted xlucene query
     */
    _restrict(q, options = {}) {
        let parser;
        const parserOptions = {
            logger: this.logger,
            type_config: this.typeConfig,
            variables: Object.assign({}, this.variables, options.variables)
        };
        try {
            parser = this._parser.make(q, parserOptions);
        }
        catch (err) {
            throw new ts.TSError(err, {
                reason: 'Query could not be parsed',
                statusCode: 422,
                context: {
                    q,
                    safe: true
                }
            });
        }
        if (p.isEmptyAST(parser.ast)) {
            if (!this.allowEmpty) {
                throw new ts.TSError('Empty queries are restricted', {
                    statusCode: 403,
                    context: {
                        q,
                        safe: true
                    }
                });
            }
            return this._addConstraints(parser, parserOptions);
        }
        parser.forTermTypes((node) => {
            // restrict when a term is specified without a field
            if (!node.field) {
                if (this.allowImplicitQueries)
                    return;
                throw new ts.TSError('Implicit fields are restricted, please specify the field', {
                    statusCode: 403,
                    context: {
                        q,
                        safe: true
                    }
                });
            }
            if (this._isFieldRestricted(node.field)) {
                throw new ts.TSError(`Field ${node.field} in query is restricted`, {
                    statusCode: 403,
                    context: {
                        q,
                        safe: true
                    }
                });
            }
            if (p.isWildcard(node)) {
                if (this.preventPrefixWildcard && startsWithWildcard(node.value)) {
                    throw new ts.TSError("Wildcard queries of the form 'fieldname:*value' or 'fieldname:?value' in query are restricted", {
                        statusCode: 403,
                        context: {
                            q,
                            safe: true
                        }
                    });
                }
            }
        });
        return this._addConstraints(parser, parserOptions);
    }
    _restrictTypeConfig() {
        const parsedConfig = {};
        for (const [typeField, value] of Object.entries(this.typeConfig)) {
            const excluded = this.excludes.filter((restrictField) => matchTypeField(typeField, restrictField));
            if (excluded.length)
                continue;
            if (this.includes.length) {
                const included = this.includes.filter((restrictField) => matchTypeField(typeField, restrictField));
                if (!included.length)
                    continue;
            }
            parsedConfig[typeField] = value;
        }
        return parsedConfig;
    }
    /**
     * Converts a restricted xlucene query to an elasticsearch search query
     *
     * @returns a restricted elasticsearch search query
     */
    async restrictSearchQuery(query, opts = {}) {
        const { params: _params = {}, variables = {}, elasticsearch_version: esVersion = 6, ...translateOptions } = opts;
        if (_params._source) {
            throw new ts.TSError('Cannot include _source in params, use _sourceInclude or _sourceExclude');
        }
        const params = { ..._params };
        const parser = this._restrict(query, { variables });
        await ts.pImmediate();
        const translator = this._translator.make(parser, {
            type_config: this.parsedTypeConfig,
            logger: this.logger,
            default_geo_field: this.defaultGeoField,
            default_geo_sort_order: this.defaultGeoSortOrder,
            default_geo_sort_unit: this.defaultGeoSortUnit,
            variables: parser.variables
        });
        const translated = translator.toElasticsearchDSL(translateOptions);
        const { includes, excludes } = this.restrictSourceFields(params._sourceInclude, params._sourceExclude);
        delete params._sourceInclude;
        delete params._sourceExclude;
        const excludesKey = esVersion >= 7 ? '_sourceExcludes' : '_sourceExclude';
        const includesKey = esVersion >= 7 ? '_sourceIncludes' : '_sourceInclude';
        const searchParams = {
            ...params,
            body: { ...params.body, ...translated },
            [excludesKey]: excludes,
            [includesKey]: includes,
        };
        if (searchParams != null) {
            delete searchParams.q;
        }
        return searchParams;
    }
    /**
     * Restrict requested source to all or subset of the ones available
     *
     * **NOTE:** this will remove restricted fields and will not throw
     */
    restrictSourceFields(includes, excludes) {
        const all = Object.keys(this.parsedTypeConfig)
            .map((field) => field.split('.')[0]);
        return {
            includes: this._getSourceFields(this.includes, all, includes),
            excludes: this._getSourceFields(this.excludes, all, excludes),
        };
    }
    _getSourceFields(restricted, all, override) {
        const fields = ts.uniq(ts.parseList(override));
        if (fields.length) {
            if (restricted.length) {
                return restricted.filter((field) => fields.includes(field));
            }
            if (all.length) {
                return fields.filter((field) => all.includes(field));
            }
            return fields;
        }
        return restricted.slice();
    }
    _isFieldRestricted(field) {
        return !Object.entries(this.parsedTypeConfig).some(([typeField, fieldType]) => {
            if (fieldType === interfaces_1.FieldType.Object)
                return false;
            const parts = typeField.split('.');
            if (parts.length > 1) {
                const firstPart = parts.slice(0, -1).join('.');
                if (this.typeConfig[firstPart] === interfaces_1.FieldType.Object) {
                    return matchFieldObject(typeField, field);
                }
            }
            return matchField(typeField, field);
        });
    }
    _addConstraints(parser, options) {
        var _a;
        if ((_a = this.constraints) === null || _a === void 0 ? void 0 : _a.length) {
            const queries = ts.concat(this.constraints, [parser.query]).filter(Boolean);
            if (queries.length === 1)
                return this._parser.make(queries[0], options);
            return this._parser.make(`(${queries.join(') AND (')})`, options);
        }
        return parser;
    }
}
exports.QueryAccess = QueryAccess;
function matchFieldObject(typeField, field) {
    const wildcardQuery = string_1.parseWildCard(field).replace(/\$$/, '');
    let s = '';
    for (const part of typeField.split('.')) {
        s += part;
        if (string_1.matchString(s, wildcardQuery)) {
            return true;
        }
        s += '.';
    }
    return false;
}
function matchField(typeField, field) {
    let s = '';
    for (const part of field.split('.')) {
        s += part;
        const wildcardQuery = string_1.parseWildCard(s);
        if (string_1.matchString(typeField, wildcardQuery)) {
            return true;
        }
        s += '.';
    }
    return false;
}
function matchTypeField(typeField, restrictField) {
    let s = '';
    for (const part of typeField.split('.')) {
        s += part;
        if (s === restrictField) {
            return true;
        }
        s += '.';
    }
    return false;
}
function startsWithWildcard(input) {
    if (!input)
        return false;
    if (!ts.isString(input))
        return false;
    return ['*', '?'].includes(ts.getFirstChar(input));
}
//# sourceMappingURL=query-access.js.map