"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@terascope/utils");
const p = __importStar(require("../parser"));
const utils_2 = require("../utils");
const string_1 = require("../document-matcher/logic-builder/string");
function translateQuery(parser, options) {
    const { logger, type_config: typeConfig } = options;
    let sort;
    function buildAnyQuery(node) {
        // if no field and is wildcard
        if (p.isWildcard(node) && !node.field && node.value === '*') {
            return {
                bool: {
                    filter: [],
                },
            };
        }
        if (p.isGroupLike(node)) {
            return buildBoolQuery(node);
        }
        if (p.isNegation(node)) {
            return buildNegationQuery(node);
        }
        if (p.isExists(node)) {
            return buildExistsQuery(node);
        }
        if (p.isTermType(node)) {
            const query = buildTermLevelQuery(node);
            if (query)
                return query;
        }
        const error = new utils_1.TSError(`Unexpected problem when translating xlucene query ${parser.query}`, {
            context: {
                node,
                ast: parser.ast,
            },
        });
        logger.error(error);
    }
    function buildTermLevelQuery(node) {
        if (p.isWildcardField(node)) {
            if (utils_1.isEmpty(typeConfig))
                throw new utils_1.TSError(`Configuration for type_config needs to be provided with fields related to ${node.field}`);
            const regex = string_1.parseWildCard(node.field);
            const should = Object.keys(typeConfig)
                .filter((field) => string_1.matchString(field, regex))
                .map((field) => Object.assign({}, node, { field }))
                .map((newNode) => buildTermLevelQuery(newNode));
            return {
                bool: {
                    should
                }
            };
        }
        if (p.isTerm(node)) {
            return buildTermQuery(node);
        }
        if (p.isRegexp(node)) {
            return buildRegExprQuery(node);
        }
        if (p.isWildcard(node)) {
            return buildWildcardQuery(node);
        }
        if (p.isRange(node)) {
            return buildRangeQuery(node);
        }
        if (p.isGeoBoundingBox(node)) {
            return buildGeoBoundingBoxQuery(node);
        }
        if (p.isGeoDistance(node)) {
            return buildGeoDistanceQuery(node);
        }
        if (p.isFunctionExpression(node)) {
            const { query, sort: sortQuery } = node.instance.toElasticsearchQuery(getTermField(node), options);
            // TODO: review how sort works in this module
            if (sortQuery != null) {
                if (!sort) {
                    sort = sortQuery;
                }
                else if (Array.isArray(sort)) {
                    sort.push(sortQuery);
                }
                else {
                    sort = [sort, sortQuery];
                }
            }
            return query;
        }
    }
    function buildMultiMatchQuery(node, query) {
        const multiMatchQuery = {
            multi_match: {
                query,
            },
        };
        logger.trace('built mutli-match query', { node, multiMatchQuery });
        return multiMatchQuery;
    }
    function buildGeoBoundingBoxQuery(node) {
        if (isMultiMatch(node))
            return;
        const field = getTermField(node);
        const geoQuery = {};
        geoQuery.geo_bounding_box = {};
        geoQuery.geo_bounding_box[field] = {
            top_left: node.top_left,
            bottom_right: node.bottom_right,
        };
        logger.trace('built geo bounding box query', { node, geoQuery });
        return geoQuery;
    }
    function buildGeoDistanceQuery(node) {
        if (isMultiMatch(node))
            return;
        const field = getTermField(node);
        const unit = node.unit || options.geo_sort_unit;
        const order = options.geo_sort_order;
        const geoQuery = {};
        geoQuery.geo_distance = {
            distance: `${node.distance}${unit}`,
        };
        geoQuery.geo_distance[field] = {
            lat: node.lat,
            lon: node.lon,
        };
        const geoQuerySort = {
            _geo_distance: {
                order,
                unit,
                [field]: {
                    lat: node.lat,
                    lon: node.lon
                }
            }
        };
        if (!sort) {
            sort = geoQuerySort;
        }
        else if (Array.isArray(sort)) {
            sort.push(geoQuerySort);
        }
        else {
            sort = [sort, geoQuerySort];
        }
        logger.trace('built geo distance query', { node, geoQuery });
        return geoQuery;
    }
    function buildRangeQuery(node) {
        if (isMultiMatch(node)) {
            if (!node.right) {
                return;
            }
            const query = `${node.left.operator}${node.left.value}`;
            return buildMultiMatchQuery(node, query);
        }
        const field = getTermField(node);
        const rangeQuery = {
            range: {
                [field]: utils_2.parseRange(node, true),
            },
        };
        logger.trace('built range query', { node, rangeQuery });
        return rangeQuery;
    }
    function buildTermQuery(node) {
        if (isMultiMatch(node)) {
            const query = `${node.value}`;
            return buildMultiMatchQuery(node, query);
        }
        const field = getTermField(node);
        if (utils_1.isString(node.value)) {
            const matchQuery = {
                match: {
                    [field]: {
                        operator: 'and',
                        query: node.value,
                    },
                },
            };
            logger.trace('built match query', { node, matchQuery });
            return matchQuery;
        }
        const termQuery = {
            term: {
                [field]: node.value,
            },
        };
        logger.trace('built term query', { node, termQuery });
        return termQuery;
    }
    function buildWildcardQuery(node) {
        if (isMultiMatch(node)) {
            const query = `${node.value}`;
            return buildMultiMatchQuery(node, query);
        }
        const field = getTermField(node);
        const wildcardQuery = {
            wildcard: {
                [field]: node.value,
            },
        };
        logger.trace('built wildcard query', { node, wildcardQuery });
        return wildcardQuery;
    }
    function buildRegExprQuery(node) {
        if (isMultiMatch(node)) {
            const query = `${node.value}`;
            return buildMultiMatchQuery(node, query);
        }
        const field = getTermField(node);
        if (node.tokenizer) {
            return {
                query_string: {
                    fields: [field],
                    query: `/${node.value}/`
                }
            };
        }
        const regexQuery = {
            regexp: {
                [field]: node.value,
            },
        };
        logger.trace('built regexpr query', { node, regexQuery });
        return regexQuery;
    }
    function buildExistsQuery(node) {
        const existsQuery = {
            exists: {
                field: node.field,
            },
        };
        logger.trace('built exists query', { node, existsQuery });
        return existsQuery;
    }
    function buildBoolQuery(node) {
        const should = [];
        for (const conj of node.flow) {
            const query = buildConjunctionQuery(conj);
            should.push(...flattenQuery(query, 'should'));
        }
        if (!should.length)
            return;
        const boolQuery = {
            bool: {
                should,
            },
        };
        logger.trace('built bool query', { node, boolQuery });
        return boolQuery;
    }
    function buildConjunctionQuery(conj) {
        const filter = [];
        for (const node of conj.nodes) {
            const query = buildAnyQuery(node);
            filter.push(...flattenQuery(query, 'filter'));
        }
        return {
            bool: {
                filter,
            },
        };
    }
    function buildNegationQuery(node) {
        const query = buildAnyQuery(node.node);
        if (!query)
            return;
        const mustNot = flattenQuery(query, 'must_not');
        logger.trace('built negation query', mustNot, node);
        return {
            bool: {
                must_not: mustNot,
            },
        };
    }
    let topLevelQuery;
    if (p.isEmptyAST(parser.ast)) {
        topLevelQuery = {
            match_all: {},
        };
    }
    else {
        const anyQuery = buildAnyQuery(parser.ast);
        const filter = compactFinalQuery(anyQuery);
        topLevelQuery = {
            constant_score: {
                filter,
            }
        };
    }
    if (!sort && options.default_geo_field && options.geo_sort_point) {
        sort = {
            _geo_distance: {
                order: options.geo_sort_order,
                unit: options.geo_sort_unit,
                [options.default_geo_field]: options.geo_sort_point,
            }
        };
    }
    return {
        query: topLevelQuery,
        // avoid setting it to undefined
        ...(sort && { sort })
    };
}
exports.translateQuery = translateQuery;
function isMultiMatch(node) {
    return !node.field || node.field === '*';
}
exports.isMultiMatch = isMultiMatch;
function getTermField(node) {
    return node.field;
}
exports.getTermField = getTermField;
function flattenQuery(query, flattenTo) {
    if (!query)
        return [];
    if (isBoolQuery(query) && canFlattenBoolQuery(query, flattenTo)) {
        return query.bool[flattenTo];
    }
    return [query];
}
exports.flattenQuery = flattenQuery;
/** This prevents double nested queries that do the same thing */
function canFlattenBoolQuery(query, flattenTo) {
    const types = Object.keys(query.bool);
    if (types.length !== 1)
        return false;
    return types[0] === flattenTo;
}
exports.canFlattenBoolQuery = canFlattenBoolQuery;
function isBoolQuery(query) {
    return query && query.bool != null;
}
exports.isBoolQuery = isBoolQuery;
function compactFinalQuery(query) {
    if (!query)
        return [];
    if (isBoolQuery(query) && canFlattenBoolQuery(query, 'filter')) {
        const filter = query.bool.filter;
        if (!filter.length)
            return query;
        if (filter.length === 1) {
            return filter[0];
        }
        return filter;
    }
    return query;
}
exports.compactFinalQuery = compactFinalQuery;
//# sourceMappingURL=utils.js.map