"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const latlon_geohash_1 = __importDefault(require("latlon-geohash"));
const utils_1 = require("@terascope/utils");
const interfaces_1 = require("./interfaces");
const helpers_1 = require("./parser/functions/geo/helpers");
function isInfiniteValue(input) {
    return input === '*' || input === Number.NEGATIVE_INFINITY || input === Number.POSITIVE_INFINITY;
}
exports.isInfiniteValue = isInfiniteValue;
function isInfiniteMin(min) {
    if (min == null)
        return false;
    return min === '*' || min === Number.NEGATIVE_INFINITY;
}
exports.isInfiniteMin = isInfiniteMin;
function isInfiniteMax(max) {
    if (max == null)
        return false;
    return max === '*' || max === Number.POSITIVE_INFINITY;
}
exports.isInfiniteMax = isInfiniteMax;
function parseRange(node, excludeInfinite = false) {
    const results = {};
    if (!excludeInfinite || !isInfiniteValue(node.left.value)) {
        results[node.left.operator] = node.left.value;
    }
    if (node.right) {
        if (!excludeInfinite || !isInfiniteValue(node.right.value)) {
            results[node.right.operator] = node.right.value;
        }
    }
    return results;
}
exports.parseRange = parseRange;
function isGreaterThan(node) {
    if (node.operator === 'gte' || node.operator === 'gt')
        return true;
    return false;
}
function buildRangeQueryString(node) {
    if (node.right) {
        const leftBrace = node.left.operator === 'gte' ? '[' : '{';
        const rightBrace = node.right.operator === 'lte' ? ']' : '}';
        return `${leftBrace}${node.left.value} TO ${node.right.value}${rightBrace}`;
    }
    // cannot have a single value infinity range query
    if (isInfiniteValue(node.left.value))
        return;
    // queryString cannot use ranges like >=1000, must convert to equivalent [1000 TO *]
    if (isGreaterThan(node.left)) {
        if (node.left.operator === 'gte') {
            return `[${node.left.value} TO *]`;
        }
        return `{${node.left.value} TO *]`;
    }
    if (node.left.operator === 'lte') {
        return `[* TO ${node.left.value}]`;
    }
    return `[* TO ${node.left.value}}`;
}
exports.buildRangeQueryString = buildRangeQueryString;
exports.GEO_DISTANCE_UNITS = {
    mi: 'miles',
    miles: 'miles',
    mile: 'miles',
    NM: 'nauticalmiles',
    nmi: 'nauticalmiles',
    nauticalmile: 'nauticalmiles',
    nauticalmiles: 'nauticalmiles',
    in: 'inch',
    inch: 'inch',
    inches: 'inch',
    yd: 'yards',
    yard: 'yards',
    yards: 'yards',
    m: 'meters',
    meter: 'meters',
    meters: 'meters',
    km: 'kilometers',
    kilometer: 'kilometers',
    kilometers: 'kilometers',
    mm: 'millimeters',
    millimeter: 'millimeters',
    millimeters: 'millimeters',
    cm: 'centimeters',
    centimeter: 'centimeters',
    centimeters: 'centimeters',
    ft: 'feet',
    feet: 'feet',
};
function parseGeoDistance(str) {
    const matches = utils_1.trim(str).match(/(\d+)(.*)$/);
    if (!matches || !matches.length) {
        throw new Error(`Incorrect geo distance parameter provided: ${str}`);
    }
    const distance = Number(matches[1]);
    const unit = parseGeoDistanceUnit(matches[2]);
    return { distance, unit };
}
exports.parseGeoDistance = parseGeoDistance;
function parseGeoDistanceUnit(input) {
    const unit = exports.GEO_DISTANCE_UNITS[utils_1.trim(input)];
    if (!unit) {
        throw new Error(`Incorrect distance unit provided: ${input}`);
    }
    return unit;
}
exports.parseGeoDistanceUnit = parseGeoDistanceUnit;
/** @returns {[lat, lon]} */
function getLonAndLat(input, throwInvalid = true) {
    let lat = input.lat || input.latitude;
    let lon = input.lon || input.longitude;
    if (helpers_1.isGeoShapePoint(input)) {
        [lon, lat] = input.coordinates;
    }
    if (throwInvalid && (!lat || !lon)) {
        throw new Error('Invalid geopoint object, it must contain keys lat,lon or latitude/longitude');
    }
    lat = utils_1.toNumber(lat);
    lon = utils_1.toNumber(lon);
    if (!utils_1.isNumber(lat) || !utils_1.isNumber(lon)) {
        if (throwInvalid)
            throw new Error('Invalid geopoint, lat and lon must be numbers');
        return null;
    }
    return [lat, lon];
}
exports.getLonAndLat = getLonAndLat;
function parseGeoPoint(point, throwInvalid = true) {
    let lat;
    let lon;
    if (typeof point === 'string') {
        if (point.match(',')) {
            [lat, lon] = utils_1.parseNumberList(point);
        }
        else {
            try {
                [lat, lon] = Object.values(latlon_geohash_1.default.decode(point));
            }
            catch (err) {
                // do nothing
            }
        }
    }
    else if (Array.isArray(point)) {
        // array of points are meant to be lon/lat format
        [lon, lat] = utils_1.parseNumberList(point);
    }
    else if (utils_1.isPlainObject(point)) {
        const results = getLonAndLat(point, throwInvalid);
        if (results)
            [lat, lon] = results;
    }
    if (throwInvalid && (lat == null || lon == null)) {
        throw new utils_1.TSError(`Invalid geopoint given to parse, point:${point}`);
    }
    // data incoming is lat,lon and we must return lon,lat
    if (lat != null && lon != null) {
        return {
            lat,
            lon
        };
    }
    return null;
}
exports.parseGeoPoint = parseGeoPoint;
const relationList = Object.values(interfaces_1.GeoShapeRelation);
function makeXluceneGeoDistanceQuery(variableState, field, value, fieldParam) {
    const dValue = fieldParam || '100m';
    const results = parseGeoPoint(value, false);
    if (!results)
        return '';
    const vDistance = variableState.createVariable('distance', dValue);
    const vPoint = variableState.createVariable('point', `${results.lat},${results.lon}`);
    return `${field}:geoDistance(point: ${vPoint}, distance: ${vDistance})`;
}
function makeXlucenePolyContainsPoint(variableState, field, value) {
    const results = parseGeoPoint(value, false);
    if (!results)
        return '';
    const vPoint = variableState.createVariable('point', `${results.lat},${results.lon}`);
    return `${field}:geoContainsPoint(point: ${vPoint})`;
}
function coordinateToXlucene(cord) {
    // tuple is [lon, lat], need to return "lat, lon"
    return `"${cord[1]}, ${cord[0]}"`;
}
exports.coordinateToXlucene = coordinateToXlucene;
function geoPolyQuery(field, polyVariableName, vParam) {
    if (vParam) {
        return `${field}:geoPolygon(points: ${polyVariableName}, relation: ${vParam})`;
    }
    return `${field}:geoPolygon(points: ${polyVariableName})`;
}
function makeXlucenePolyQuery(variableState, field, polyVariableName, fieldParam) {
    let vParam;
    if (fieldParam && relationList.includes(fieldParam)) {
        vParam = variableState.createVariable('relation', fieldParam);
    }
    return geoPolyQuery(field, polyVariableName, vParam);
}
function createGeoQuery(variableState, field, value, targetType, fieldParam) {
    if (helpers_1.isGeoJSONData(value)) {
        if (helpers_1.isGeoShapePolygon(value) || helpers_1.isGeoShapeMultiPolygon(value)) {
            // geoPolygon internally takes care of poly/multipoly.
            // need to preserve original poly in variables
            const polyVariableName = variableState.createVariable('points', value);
            return makeXlucenePolyQuery(variableState, field, polyVariableName, fieldParam);
        }
        if (helpers_1.isGeoShapePoint(value)) {
            if (isGeoPointType(targetType)) {
                return makeXluceneGeoDistanceQuery(variableState, field, value.coordinates, fieldParam);
            }
            return makeXlucenePolyContainsPoint(variableState, field, value.coordinates);
        }
        // We do not support any other geoJSON types;
        return '';
    }
    // incoming value is a geo-point and we compare to another geo-point by geoDistance query
    if (isGeoPointType(targetType)) {
        return makeXluceneGeoDistanceQuery(variableState, field, value, fieldParam);
    }
    if (isGeoJSONType(targetType))
        return makeXlucenePolyContainsPoint(variableState, field, value);
    // if here then return a noop
    return '';
}
function isGeoQuery(type) {
    return isGeoPointType(type) || isGeoJSONType(type);
}
function isGeoPointType(type) {
    return type === interfaces_1.FieldType.Geo || type === interfaces_1.FieldType.GeoPoint;
}
function isGeoJSONType(type) {
    return type === interfaces_1.FieldType.GeoJSON;
}
// TODO: handle variables and query from before
function createJoinQuery(input, options = {}) {
    const { fieldParams = {}, joinBy = 'AND', typeConfig = {}, variables } = options;
    const variableState = new VariableState(variables);
    const query = Object.entries(input)
        .map(([field, val]) => {
        if (val == null)
            return '';
        const config = typeConfig[field];
        if (isGeoQuery(config)) {
            return createGeoQuery(variableState, field, val, config, fieldParams[field]);
        }
        const value = variableState.createVariable(field, val);
        return `${field}: ${value}`;
    })
        .filter(Boolean)
        .join(` ${joinBy} `)
        .trim();
    const finalVariables = variableState.getVariables();
    return { query, variables: finalVariables };
}
exports.createJoinQuery = createJoinQuery;
class VariableState {
    constructor(variables) {
        this.variables = { ...variables };
    }
    _makeKey(field) {
        let num = 1;
        let name = `${field}_${num}`;
        while (this.variables[name] !== undefined) {
            num += 1;
            name = `${field}_${num}`;
        }
        return name;
    }
    createVariable(field, value) {
        if (typeof value === 'string' && utils_1.startsWith(value, '$')) {
            const vField = value.slice(1);
            if (this.variables[vField] === undefined)
                throw new Error(`Must provide variable "${vField}" in the variables config`);
            return value;
        }
        const key = this._makeKey(field);
        this.variables[key] = value;
        return `$${key}`;
    }
    /**
     * Shallow clones and sorts the keys
    */
    getVariables() {
        const result = {};
        for (const key of Object.keys(this.variables).sort()) {
            result[key] = this.variables[key];
        }
        return result;
    }
}
exports.VariableState = VariableState;
//# sourceMappingURL=utils.js.map